/*
 * Very hoky program for playing with pcr-1000
 * Usage: prog [-c] [-t tty_device_name]
 *       -c to use curses (defaults to readline version)
 *       -t to set the filename of the device to open.
 */

#include <stdlib.h>
#include <stdio.h>
#include <curses.h>
#include <stdarg.h>
#include <assert.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <libgen.h>
#include <ctype.h>

#include "ttyio.h"
#include "xlater.h"
#include <readline.h>

/*
 * These odd pointers are so that both curses and readline versions
 * of the program can coexist.
 */

int use_curses = 0;
void *output;
void *cmd_output;
void *cmd_input;
void *pcr_output;
void *pcr_input;

static WINDOW *pcr_in_window;	/* debug raw text from pcr */
static WINDOW *cmd_in_window;	/* debug nearly raw text from keyboard */
static WINDOW *pcr_out_window;	/* cooked text from pcr */
static WINDOW *cmd_out_window;	/* cmds as sent to pcr */

/* The only command that requires the program to do something
 * besides io is setting the baud rate.
 * This variable gets set in the command recognizer, so that
 * the program can try to actually set the baud rate on the
 * interface.  Once set, the program should make this back to -1
 * it is set to the index of the baud rate in the baud rate table
 */
static int setting_baudrate = -1;

static int baudrates[6] = {B1200, B2400, B4800, B9600, B19200, B38400};

/*
 * inbuf is data from pcr-1000, outbuf is data from keyboard
 */

#define BUFSIZE 80
static char inbuf[BUFSIZE];
static char outbuf[BUFSIZE];

static int ocount = 0;

#define CR 13
#define LF 10

static const char * const default_tty_name = "/dev/tty.USA19QI12P1.1";

/* A static variable for holding the line. */
static char *line_read = (char *)NULL;

/* Read a string, and return a pointer to it.
   Returns NULL on EOF. */
char *
rl_gets (const char * const prompt)
{
  /* If the buffer has already been allocated,
     return the memory to the free pool. */
  if (line_read)
    {
      free (line_read);
      line_read = (char *)NULL;
    }

  /* Get a line from the user. */
  line_read = readline (prompt);

  /* If the line has any text in it,
     save it on the history. */
  if (line_read && *line_read)
    add_history (line_read);

  return (line_read);
}

void add_in_char(const char * const b, int c);
void add_stdio_cmd_char(int c);
void add_curses_cmd_char(int c);
void add_cmd_char(int c);

char buffer[1023];

int  tty_poll(void) {
  int j;

  /* get any characters available from the radio */
  if ((j = get_bytes_pcr1000(&inbuf[0], BUFSIZ)) > 0) {
    if (use_curses) {
	waddnstr(pcr_in_window, inbuf, j);
	wrefresh(pcr_in_window);
    }
    add_in_char(inbuf, j);
  }
  return 0;
}

void myprintf(void *dest, char *fmt, ...)
{

  va_list ap;
  va_start(ap, fmt);

  if (use_curses) {
    vw_printw((WINDOW *)dest, fmt, ap);
  } else {
    vfprintf((FILE *)dest, fmt, ap);
  }

}

void init_stdio()
{

  output = (void *)stdout;

  cmd_output = cmd_input = pcr_output = pcr_input = output;

  rl_event_hook = tty_poll;

}

void init_curses()
{
  int i, j, k, l;
  int c;
  int maxy, maxx;
  
  /* set up curses */
  initscr(); cbreak(); noecho(); nodelay(stdscr, TRUE);
  nonl(); intrflush(stdscr, FALSE);

  getmaxyx(stdscr, maxy, maxx);

  /* The screen is divided into four regions. the top 1/2 of the screen
   * is used for raw input from the radio debugging. The next 1/4 of the
   * screen is used for formatted radio data.
   * The remaining quarter is divided so that data input happens on the left
   * and formatted commands/queries are displayed on the right.
   */

  i = maxy/4;
  j = 0;
  k = maxx/2;
  pcr_in_window = newwin(i, 0, 0, 0);
  scrollok(pcr_in_window, TRUE);

  j += i;
  pcr_out_window = newwin(2*i, 0, j, 0);
  scrollok(pcr_out_window, TRUE);

  j += 2*i;
  cmd_in_window = newwin(i, k, j, 0);
  scrollok(cmd_in_window, TRUE);
  nodelay(cmd_in_window, TRUE);

  cmd_out_window = newwin(i, k, j, k);
  scrollok(cmd_out_window, TRUE);

  cmd_output = (void *)cmd_out_window;
  cmd_input = (void *)cmd_in_window;
  pcr_output = (void *)pcr_out_window;
  pcr_input = (void *)pcr_in_window;
}

int main(int argc, char *argv[])
{
  int i, j, k, l;
  int c;
  char * tty_name;
  char * prog_name;
  char *cmd;
  char *x;
  char saved_mode[3];
  char saved_filter[3];
  int have_mode = 0;
  int have_filter = 0;

  tty_name = strdup(default_tty_name);
  prog_name = basename(argv[0]);

  /* crack args */
  while ((c = getopt(argc, argv, "ct:")) != -1) {
    switch (c) {
    case 'c':
      use_curses = 1;
      break;
    case 't':
      tty_name = optarg;
      break;
    case '?':
    default:
      fprintf(stderr,"*Usage: %s [-c] [-t ttyname]\n", prog_name);
      return 1;
    }
  }

  argc -= optind;
  argv += optind;

  if (argc != 0) {
    myprintf(stderr,"Usage: %s [tty name]\n", prog_name);
    return 1;
  }

  if (open_pcr1000(tty_name) < 0)
      return 1;

  if (use_curses) {

    init_curses();

    while (c = wgetch(cmd_in_window)) {
      /* our exit key(s) */
      if (c == '`' || c == '~' || c == 'Q' || c == 'q')
	break;
      if (c != ERR) {		/* process it */
	add_cmd_char(c);
      }
      tty_poll();
    }

    endwin();

  } else {

    init_stdio();

    while ( 1 ) {
      cmd = rl_gets("> ");
      c = *cmd;

      /* our exit key(s) */
      if (c == '`' || c == '~' || c == 'Q' || c == 'q')
	break;

      x = xlate(cmd);

      if (x) {
	myprintf(output,"xlate(%s) = [%s]\n", cmd, x);
	if (*x == 'm') {		/* special case for mode */
	  saved_mode[0] = x[1];
	  saved_mode[1] = x[2];
	  saved_mode[2] = 0;
	  have_mode = 1;
	  myprintf(output,"Saving mode %s\n", saved_mode);
	} else if (*x == 'f') {
	  saved_filter[0] = x[1];
	  saved_filter[1] = x[2];
	  saved_filter[2] = 0;
	  have_filter = 1;
	  myprintf(output,"Saving filter %s\n", saved_filter);
	} else if (*x == 'K') {
	  if (!have_mode || !have_filter) {
	    myprintf(output,"Can't set freq until mode is set.\n");
	  } else {
	    sprintf(buffer,"%s%s%s00", x, saved_mode, saved_filter);
	    myprintf(output,"xlate(%s) = [%s]\n", cmd, buffer);
	    i = 0;
	    while (buffer[i]) {
	      add_cmd_char(buffer[i]);
	      i++;
	    }
	    add_cmd_char('.');
	  }
	} else {
	  myprintf(output,"xlate(%s) = [%s]\n", cmd, x);
	  i = 0;
	  while (x[i]) {
	    add_cmd_char(x[i]);
	    i++;
	  }
	  add_cmd_char('.');
	}
	free(x);

      } else {
	
	myprintf(output,"xlate(%s) is invalid.\n", cmd);
      }
      
    }
  }

  close_pcr1000();

  return 0;

}

void fprintnstr(void *fd, const char * const buf, int len)
{
  int i;
  for (i = 0; i < len; i++)
    myprintf(fd,"%c", buf[i]);
  return;
}

/* process the PCR's output */
/* possible returns are:
 * G00x  == last command status x = 0 is good, x = 1 is bad
 * G210  == PCR-1000 protocol
 * GD0x  == optional devices bit 0 = 1 if 106 present (bit 4 for DARC 107)
 * GExx  == Country code (see table)
 * H10x  == linear power status x = 0 is off, x = 1 is on
 * H9xx  == Scan status return code (see table)
 * I00x  == busy status (see table)
 * I1xx  == signal strength
 * I2xx  == centering meter 00 = < 80 = . FF = >
 * I31x  == dtmf tone received
 * NE1x0 == bandscope data output
 */

#define BLIND 0
#define BAND_SCOPE 1
#define ORDINARY 2
#define NOISE 3
static int state = BLIND;

int mycnt = 0;
char mybuf[80];

static int lineno = 0;
void show_buf()
{
  int i;
  lineno++;
  myprintf(pcr_output, "%2d: ", lineno);
  if (state == BAND_SCOPE) {
    myprintf(pcr_output,"bandscope data: [");
    fprintnstr(pcr_output,mybuf,mycnt);
    myprintf(pcr_output,"]\n");
    return;
  }
  if (state != ORDINARY) {
    myprintf(pcr_output,"confused: [");
    fprintnstr(pcr_output,mybuf,mycnt);
    myprintf(pcr_output,"]\n");
    return;
  }
  myprintf(pcr_output,"status: [");
  fprintnstr(pcr_output,mybuf,mycnt);
  myprintf(output,"] ");
  switch(mybuf[0]) {
  case 'G':
    switch(mybuf[1]) {
    case '0':
      myprintf(pcr_output," last command status %s\n",
	      mybuf[3] == '0' ? "good" : "bad");
      break;
    case '2':
      myprintf(pcr_output," protocol is ");
      fprintnstr(pcr_output,&mybuf[2],2);
      myprintf(pcr_output,"\n");
      break;
    case 'D':
      i = mybuf[3] - '0';
      myprintf(pcr_output," DSP %s ", i & 1 ? "present" : "absent");
      myprintf(pcr_output," DARC %s", i & (1<<4) ? "present" : "absent");
      myprintf(pcr_output,"\n");
      break;
    case 'E':
      myprintf(pcr_output," country code is ");
      fprintnstr(pcr_output,&mybuf[2],2);
      myprintf(pcr_output,"\n");
      break;
    default:
      break;
    }
    break;
  case 'H':
    switch(mybuf[1]) {
    case '1':
      myprintf(pcr_output," linear power is %s\n",
	      mybuf[3] == '0' ? "off" : "on");
      break;
    case '9':
      myprintf(pcr_output," scan status is ");
      if (mybuf[2] == '0') {
	switch(mybuf[3]) {
	case '0':
	  myprintf(pcr_output,"Not scanning\n");
	  break;
	case '1':
	  myprintf(pcr_output,"Ready to move\n");
	  break;
	case '2':
	  myprintf(pcr_output,"Waiting for squelch\n");
	  break;
	case '3':
	  myprintf(pcr_output,"reviewing squelch\n");
	  break;
	case '4':
	  myprintf(pcr_output,"evaluating tone squelch\n");
	  break;
	case '5':
	  myprintf(pcr_output,"evaluating vsc\n");
	  break;
	default:
	  break;
	}
      }	else if (mybuf[2] == '1') {
	  myprintf(pcr_output,"halt due to busy\n");
      }
      break;
    default:
      break;
    }
    break;
  case 'I':
    switch(mybuf[1]) {
    case '0':
      i = mybuf[3] - '0';
      myprintf(pcr_output,"scanner is %s", i & 1 ? "busy" : "not busy");
      if (i & 2)
	myprintf(pcr_output, ", AF open");
      if (i & 4)
	myprintf(pcr_output, ", VSC open");
      if (i & (1<<7))
	myprintf(pcr_output, ", AF receiver error");
      myprintf(pcr_output,"\n");
      break;
    case '1':
      myprintf(pcr_output,"S meter = ");
      fprintnstr(pcr_output,&mybuf[2],2);
      myprintf(pcr_output,"\n");
      break;
    case '2':
      myprintf(pcr_output,"centering = ");
      fprintnstr(pcr_output,&mybuf[2],2);
      myprintf(pcr_output,"\n");
      break;
    case '3':
      myprintf(pcr_output,"dtmf = %c\n", mybuf[3]);
      break;
    default:
      break;
    }
    break; 
  default:
    break;
  }
}

void add_in_char(const char * const b, int c)
{
  int i;
  int x;
  for (i = 0; i < c; i++) {
    x = b[i];
    switch (x) {
    case CR:			/* assume the next char is LF */
      if (mycnt)
	show_buf();
      mycnt = 0;
      state = BLIND;
      break;
    case LF:			/* assume the prev char was CR */
      if (mycnt)
	show_buf();
      mycnt = 0;
      state = BLIND;
      break;
    case 'G':
    case 'H':
    case 'I':
      if (mycnt)
	show_buf();
      mybuf[0] = x;
      mycnt = 1;
      state = ORDINARY;
      break;
    case 'N':
      if (mycnt)
	show_buf();
      mybuf[0] = x;
      mycnt = 1;
      state = BAND_SCOPE;
      break;
    default:			/* this could be tighter */
      mybuf[mycnt++] = x;
      if ((state == ORDINARY && mycnt == 4)
	  || (state == BAND_SCOPE && mycnt == 37)) {
	show_buf();
	mycnt = 0;
	state = BLIND;
      }
      break;
    }
  }
}

/* valid query:
 * G0? -- ask for status
 * G2? -- ask for protocol version
 * GD? -- ask about optional devices
 * GE? -- ask about destination country
 * H1? -- ask if linear power is on
 * H9? -- ask about scan status
 * I0? -- ask if busy
 * I1? -- ask for s meter level
 * I2? -- ask for center meter level
 * I3? -- ask for DTMF decode
 * NE1x0? -- ask for bandscope output
 */
int is_valid_query()
{
  int prefix = outbuf[0];
  int cmd = outbuf[1];
  int q = outbuf[2];
  switch (prefix) {
  case 'G':
    if (ocount != 3)
      return 0;
    switch (cmd) {
    case '0':
      myprintf(cmd_output," status ");
      return q == '?';
    case '2':
      myprintf(cmd_output," protocol version ");
      return q == '?';
    case 'D':
      myprintf(cmd_output," optional devices ");
      return q == '?';
    case 'E':
      myprintf(cmd_output," destination country ");
      return q == '?';
    default:
      myprintf(cmd_output," unknown ");
      return 0;
    }
  case 'H':
    if (ocount != 3)
      return 0;
    switch (cmd) {
    case '1':
      myprintf(cmd_output," linear power ");
      return q == '?';
    case '9':
      myprintf(cmd_output," scan status ");
      return q == '?';
    default:
      myprintf(cmd_output," unknown ");
      return 0;
    }
  case 'I':
    if (ocount != 3)
      return 0;
    switch (cmd) {
    case '0':
      myprintf(cmd_output," busy ");
      return q == '?';
    case '1':
      myprintf(cmd_output," s meter level ");
      return q == '?';
    case '2':
      myprintf(cmd_output," center meter level ");
      return q == '?';
    case '3':
      myprintf(cmd_output," DTMF decode ");
      return q == '?';
    default:
      myprintf(cmd_output," unknown ");
      return 0;
    }
  case 'N':
    if (ocount != 6)
      return 0;
    if (cmd != 'E')
      return 0;
    if (q != '1')
      return 0;
    if (!isxdigit(outbuf[3]) || outbuf[4] != '0' ||  outbuf[5] != '?')
      return 0;
    myprintf(cmd_output," bandscope data %c", outbuf[3]);
    return 1;
  default:
    return 0;
  }
  return 0;
}

/*
 * yuck. command validation the hard way.
 *
 */
int is_valid_zero(int p)
{
  return (outbuf[p] == '0' && outbuf[p+1] == '0');
}

int is_valid_onoff(int p)
{
  return (outbuf[p] == '0' && (outbuf[p+1] == '0' || outbuf[p+1] == '1'));
}

int is_valid_baudrate(int p)
{
  if (outbuf[p] != '0')
    return 0;
  switch (outbuf[p+1]) {
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
    return 1;
  default:
    return 0;
  }
  return 0;
}

int is_valid_frequency(int p)
{
  int i;
  for (i = p; i < p + 10; i++)
    if (!isdigit(outbuf[i]))
      return 0;
  return 1;
}

int is_valid_mode(int p)
{
  if (outbuf[p] != '0')
    return 0;
  switch (outbuf[p+1]) {
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
    return 1;
  default:
    return 0;
  }
  return 0;
}

int is_valid_filter(int p)
{
  if (outbuf[p] != '0')
    return 0;
  switch (outbuf[p+1]) {
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
    return 1;
  default:
    return 0;
  }
  return 0;
}

int is_valid_digits(int p)
{
  return (isdigit(outbuf[p]) && isdigit(outbuf[p+1]));
}

int is_valid_hex_digits(int p)
{
  return (isxdigit(outbuf[p]) && isxdigit(outbuf[p+1]));
}

int is_valid_command()
{
  char prefix = outbuf[0];
  char command = outbuf[1];

  /* all commands are at least 4 bytes long */
  if (ocount < 4)
    return 0;

  switch (prefix) {
  case 'G':
  case 'H':
    if (ocount != 4)
      return 0;
    break;
  case 'J':
    if (ocount != 5)
      return 0;
    break;
  case 'K':
    if (ocount != 18)
      return 0;
    break;
#ifdef LATER
  case 'L':
    if (ocount != 7)
      return 0;
    break;
#endif
  case 'M':
    if (ocount != 16)
      return 0;
    break;
  default:
    return 0;
  }

  /* Now for the messy details */
  switch (prefix) {
  case 'G':
    switch(command) {
    case '1':			/* baud rate */
      if (is_valid_baudrate(2)) {
	setting_baudrate = outbuf[3] - '0';
	return 1;
      }
      return 0;
    case '3':			/* communication mode */
      return is_valid_onoff(2);
    default:
      return 0;
    }
  case 'H':
    switch(command) {
    case '0':			/* software reset */
      return is_valid_zero(2);
    case '1':			/* linear power on/off */
      return is_valid_onoff(2);
    case '8':			/* how to scan */
      return is_valid_digits(2);
    default:
      return 0;
    }
  case 'J':
    switch(command) {
    case '4':
      switch(outbuf[2]) {
      case '0':			/* J40 -- set volume level */
      case '1':			/* J41 -- set squelch level */
      case '3':			/* J43 -- set IF shift */
      case 'A':			/* J4A -- set BFO */
	return is_valid_hex_digits(2);
      case '5':			/* J45 -- set AGC */
      case '6':			/* J46 -- set noise blanker */
      case '7':			/* J47 -- set attenuator */
	return is_valid_onoff(2);
      default:
	return 0;
      }
    case '5':
      switch(outbuf[2]) {
      case '0':			/* j50 -- set VSC */
	return is_valid_onoff(2);
      case '1':			/* J51 -- set CTCSS (?) */
	return outbuf[3] == '0' && isxdigit(outbuf[4]);
      default:
	return 0;
      }
    case '8':
      switch(outbuf[2]) {
      case '0':			/* J80 -- set DSP prefix */
	return outbuf[3] == '0' && outbuf[4] == '1';
      case '1':			/* J81 -- enable DSP */
	return is_valid_onoff(2);
      case '2':			/* J82 -- noise reduction level */
	return is_valid_hex_digits(2);
      case '3':			/* J83 -- enable autonotch filter */
	return is_valid_onoff(2);
      default:
	return 0;
      }
    default:
      return 0;
    }
  case 'K':			/* K0 -- set freq and mode */
    return command == '0' && is_valid_frequency(2)
      && is_valid_mode(12)
      && is_valid_filter(14);
  case 'L':			/* LE200 -- ??? */
    return 0;
  case 'M':			/* set bandscope. defered */
    return 1;
  default:
    return 0;
  }

  return 0;
}

void add_cmd_char(int c)
{
  if (use_curses)
    add_curses_cmd_char(c);
  else
    add_stdio_cmd_char(c);
}

/* process keyboard input */
void add_stdio_cmd_char(int c)
{
  /* ordinary text. ignore nonalphanumerics
   * convert lower to upper
   * print it on the raw window
   * and return
   */
  if (isalnum(c)) {
    if (islower(c))
      c = toupper(c);
    outbuf[ocount++] = c;
    return;
  }

  /* queries always end with a '?'
   * if you get a query, add the '?'
   * check to see if it's a valid query
   * and send it to the radio if it is
   */
  if (c == '?') {
    outbuf[ocount++] = c;
    myprintf(cmd_output,"Query:");
    fprintnstr(cmd_output,outbuf,ocount);
    myprintf(cmd_output,"?");
    if (is_valid_query()) {
      myprintf(cmd_output," send.\n");
      outbuf[ocount++] = CR;
      outbuf[ocount++] = LF;
      put_bytes_pcr1000(outbuf, ocount);
    } else {
      myprintf(cmd_output," is not recognized.\n");
    }
    ocount = 0;
    return;
  }

  /* commands end with a '.' but it isn't sent to the radio.
   * if you get a command, see if it's valid.
   * and send it to the radio if is.
   */

  if (c == '.') {
    myprintf(cmd_output,"Command: ");
    fprintnstr(cmd_output,outbuf,ocount);
    if (is_valid_command()) {
      myprintf(cmd_output," send.\n");
      outbuf[ocount++] = CR;
      outbuf[ocount++] = LF;
      put_bytes_pcr1000(outbuf, ocount);
      if (setting_baudrate != -1) {
	set_baudrate_pcr1000(baudrates[setting_baudrate],
			     baudrates[setting_baudrate]);
	myprintf(cmd_output, "[baudrates[%d] = %d]", setting_baudrate,
		baudrates[setting_baudrate]);
	setting_baudrate = -1;
      }
      ocount = 0;
      return;
    } else {
      myprintf(cmd_output," is not recognized.\n");
    }
  }

  /* if we get here, it's a char we don't want to deal with,
   * so we throw it away without doing anything.
   */
}

/* process keyboard input */
void add_curses_cmd_char(int c)
{
  int y, x;
  /* cover terminal i/o first */
  if (c == erasechar()) {	/* delete char */
    if (ocount) {
      ocount--;
      getyx(cmd_in_window, y, x);
      x--;
      mvwdelch(cmd_in_window, y, x);
    }
    return;
  }
  if (c == killchar()) {	/* delete line */
    if (ocount) {
      ocount = 0;
      getyx(cmd_in_window, y, x);
      x = 0;
      wmove(cmd_in_window, y, x);
      wclrtoeol(cmd_in_window);
    }
    return;
  }

  /* ordinary text. ignore nonalphanumerics
   * convert lower to upper
   * print it on the raw window
   * and return
   */
  if (isalnum(c)) {
    if (islower(c))
      c = toupper(c);
    outbuf[ocount++] = c;
    waddch(cmd_in_window,c);
    wrefresh(cmd_in_window);
    return;
  }

  /* queries always end with a '?'
   * if you get a query, add the '?'
   * check to see if it's a valid query
   * and send it to the radio if it is
   */
  if (c == '?') {
    outbuf[ocount++] = c;
    waddch(cmd_in_window,c);
    waddch(cmd_in_window,'\n');
      waddstr(cmd_out_window,"QUERY (");
      waddnstr(cmd_out_window,outbuf,ocount);
      waddstr(cmd_out_window,")");
    if (is_valid_query()) {
      outbuf[ocount++] = CR;
      outbuf[ocount++] = LF;
      put_bytes_pcr1000(outbuf, ocount);
    } else {
      waddstr(cmd_out_window,"????");
    }
    waddstr(cmd_out_window,"\n");
    wrefresh(cmd_in_window);
    wrefresh(cmd_out_window);
    ocount = 0;
    return;
  }

  /* commands end with a '.' but it isn't sent to the radio.
   * if you get a command, see if it's valid.
   * and send it to the radio if is.
   */

  if (c == '.') {
    if (is_valid_command()) {
      waddstr(cmd_out_window,"COMMAND (");
      waddnstr(cmd_out_window,outbuf,ocount);
      outbuf[ocount++] = CR;
      outbuf[ocount++] = LF;
      put_bytes_pcr1000(outbuf, ocount);
      if (setting_baudrate != -1) {
	set_baudrate_pcr1000(baudrates[setting_baudrate],
			     baudrates[setting_baudrate]);
	wprintw(cmd_out_window, "[baudrates[%d] = %d]", setting_baudrate,
		baudrates[setting_baudrate]);
	setting_baudrate = -1;
      }
    } else {
      waddstr(cmd_out_window,"???? (");
      waddnstr(cmd_out_window,outbuf,ocount);
    }
    waddstr(cmd_out_window,")\n");
    wrefresh(cmd_out_window);
    waddch(cmd_in_window,'\n');
    wrefresh(cmd_in_window);
    ocount = 0;
    return;
  }

  /* if we get here, it's a char we don't want to deal with,
   * so we throw it away without doing anything.
   */
}
