#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <ctype.h>

#include "ttyio.h"
static int term_initialized = 0;
static char *term_name = 0;
static int tty_fd = -1;
static struct termios saved_term;

static int put_termios(struct termios const * const src_term)
{
  int i;
  if (ioctl(tty_fd, TIOCSETAF, src_term) < 0) {
#if 1
    fprintf(stderr,"Setting termios failed on \"%s\"\n", term_name);
    perror("ioctl");
#endif
    return -1;
  }
  return 0;
}

static int get_termios(struct termios *term)
{
  if (ioctl(tty_fd, TIOCGETA, term) < 0) {
#ifdef VERBOSE
    fprintf(stderr,"Getting termios failed on \"%s\"\n", term_name);
    perror("ioctl");
#endif
    return -1;
  }
  return 0;
}

static void setup(struct termios *term)
{
  term->c_iflag = BRKINT;	/* turn break into sigint */
  term->c_oflag &= ~OPOST;	/* do no output processing */
  term->c_cflag |=  ( CLOCAL | CREAD );
  term->c_lflag  &= ~(ICANON | ECHO | ECHOE | ISIG);
  term->c_cc[VMIN] = 0;
  term->c_cc[VTIME] = 1;	/* .1 second */
  term->c_ispeed = B9600;
  term->c_ospeed = B9600;
}

/*
 * open the tty in a mode suitable for controlling the PCR 1000
 */
int open_pcr1000(const char * const tty_name)
{
  int fd;
  struct termios term;
  fd = open(tty_name, O_RDWR | O_NOCTTY | O_NDELAY, 0);
  if (fd < 0) {
#ifdef VERBOSE
    fprintf(stderr,"Can not open \"%s\"\n", tty_name);
    perror("open");
#endif
    return -1;
  }
  tty_fd = fd;
  term_name = malloc(strlen(tty_name)+1);
  if (term_name == 0) {
#ifdef VERBOSE
    fprintf(stderr,"Can not allocate space for term name\n");
#endif
  } else {
    strcpy(term_name, tty_name);
  }

  if (fcntl(fd, F_SETFL, 0) < 0) {
#ifdef VERBOSE
    fprintf(stderr,"Can not clear O_NDELAY on \"%s\": %s (%d).\n",
	    tty_name, strerror(errno), errno);
#endif
    close_pcr1000();
    return -1;
  }
  if (get_termios(&saved_term) < 0) {
    close_pcr1000();
    return -1;
  }
  term = saved_term;
  setup(&term);
  if (put_termios(&term) < 0) {
    close_pcr1000();
    return -1;
  }
  term_initialized = 1;
  return 0;
}

int close_pcr1000()
{
  if (term_initialized) {
    put_termios(&saved_term);
    term_initialized = 0;
  }
  if (close(tty_fd) < 0) {
#ifdef VERBOSE
    fprintf(stderr,"Unexpected! Can not close \"%s\"\n", term_name);
    perror("close");
#endif
    return -1;
  }
  free(term_name);
  return 0;
}

static int get_ldisc(const char * const term_name, int tty, int *ldisc)
{
  int newldisc;
  if (ioctl(tty, TIOCGETD, &newldisc) < 0) {
#ifdef VERBOSE
    fprintf(stderr,"Getting line discipline failed on \"%s\"\n", term_name);
    perror("ioctl");
#endif
    return -1;
  }
  *ldisc = newldisc;
  return 0;
}

int set_baudrate_pcr1000(int irate, int orate)
{
  struct termios term;
  int i, o;

  if (get_termios(&term) < 0)
    return -1;
  
  i = cfsetispeed(&term, irate);
  o = cfsetospeed(&term, orate);
  if (i < 0 || o < 0)
    return -1;
  return put_termios(&term);
}

int get_baudrate_pcr1000(int *ispeed, int *ospeed)
{
  struct termios term;
  if (get_termios(&term) < 0)
    return -1;
  *ispeed = term.c_ispeed;
  *ospeed = term.c_ospeed;
  return 0;
}

int put_bytes_pcr1000(const char * const buffer, int nbytes)
{
  int i = write(tty_fd, buffer, nbytes);
  if (i == nbytes)
    return 0;
  if (i > 0) {
#ifdef VERBOSE
    fprintf(stderr,"Exepected to write %d bytes to \"%s\", wrote %d\n",
	    nbytes, term_name, i);
#endif
    return -1;
  }
#ifdef VERBOSE
  fprintf(stderr,"Error tryin to write %d bytes to \"%s\"\n",
	  term_name, nbytes);
  perror("write");
#endif
  return -1;
}

int get_bytes_pcr1000(char * buffer, int maxbytes)
{
  int i = read(tty_fd, buffer, maxbytes);
  if (i > 0)
    return i;
  if (i == 0) {
#ifdef VERBOSE
    fprintf(stderr,"EOF trying to read %d bytes  \"%s\"\n",
	    maxbytes, term_name);
#endif
    return 0;
  }
#ifdef VERBOSE
  fprintf(stderr,"Error trying to write %d bytes from \"%s\"\n",
	  maxbytes, term_name);
  perror("write");
#endif
  return -1;
}

