#include <stdlib.h>
#include <ctype.h>

#include "xlater.h"
#include "pcrdefs.h"

typedef struct _list {
  char * source;
  char * dest;
} List;

List OnOff[] = { {"on", "1"}, {"off", "0"}, {0, 0} };
List Baud[] = { {"1200", "00"}, {"2400", "01",}, {"4800", "02"},
		{"9600", "03"}, {"19200", "04",}, {"38400", "05"}, {0, 0} };
List Mode[] = { {"lsb", "00"}, {"usb", "01"}, {"am", "02"},
		{"cw", "03"}, {"nfm", "05"}, {"wfm", "06"}, {0, 0} };
List Filter[] = { {"2k8", "00"}, {"6k", "01"}, {"15k", "02"},
		  {"50k", "03"}, {"230k", "04"}, {0, 0} };

struct _command {
  char * c_name;
  char * c_prefix;
  unsigned int c_type;
  List * c_opt;
} Commands[] = {
  { "baud", "G1", TYPE_LIST, Baud },
  { "auto", "G30", TYPE_ONOFF, OnOff },
  { "receiver", "H10", TYPE_ONOFF, OnOff },
  { "volume", "J40", TYPE_RANGE, 0 },
  { "squelch", "J41", TYPE_RANGE, 0 },
  { "ifshift", "J43", TYPE_RANGE, 0 },
  { "agc", "J450", TYPE_ONOFF, OnOff },
  { "nb", "J460", TYPE_ONOFF, OnOff },
  { "att", "J470", TYPE_ONOFF, OnOff },
  { "vsc", "J500", TYPE_ONOFF, OnOff },
  { "tsq", "J51", TYPE_LIST, 0 },
  { "frequency", "K0", TYPE_FREQ, 0 },
  { "mode", "m", TYPE_LIST, Mode },
  { "filter", "f", TYPE_LIST, Filter },
  { "bandscope", "ME00001", TYPE_BAND, 0},
  { "firmware", "G4", TYPE_QUERY, 0},
  { "region", "GE", TYPE_QUERY, 0},
  { "dsp", "GD", TYPE_QUERY, 0},
  { "power", "H1", TYPE_QUERY, 0},
  { "sq", "I0", TYPE_QUERY, 0},
  { "smeter", "I1", TYPE_QUERY, 0},
  { "center", "I2", TYPE_QUERY, 0},
  { "", "", TYPE_QUERY, 0},
  { 0, 0, 0, 0 }
};


/*
 * most commands are of the form COMMAND VALUE
 * and this parser copes with that.
 */

/* A couple of temporary buffers */

#define WORD_LEN 80		/* *way* over sized buffer */
char command[WORD_LEN];
char argument[WORD_LEN];

char buffer[WORD_LEN];

int findArg(const char * const argument, const List * const list)
{
  int i;
  int l;
  int cnt = 0;
  int argNo = -1;
  int argLen = strlen(argument);
  
  for (i = 0; list[i].source; i++) {
    l = strlen(list[i].source);
    if (l > argLen) l = argLen;
    if (strncmp(argument, list[i].source, l) == 0) {
      cnt++;
      argNo = i;
    }
  }
  if (cnt == 0) {
    printf("Invalid argument %s\n", argument);
    return -1;
  } else if (cnt > 1) {
    printf("Ambiguous argument %s\n", argument);
    return -1;
  }
  return argNo;
}

int toInt(const char * const argument)
{
  char *endptr;
  long r = strtol(argument, &endptr, 0);
  if (*endptr)
    return -1;
  return r;
}

double toDouble(const char * const argument)
{
  char *endptr;
  double r = strtod(argument, &endptr);
  if (*endptr)
    return -1.0;
  return r;
}

char *xlate(const char * const cmd)
{
  int i;
  int l;
  int cnt;
  int cmdNo;
  int argNo;
  int cmdLen = 0;
  int argLen = 0;
  double freq;
  char *retval;

  /* isolate the command */
  /* skip to the first alpha char */
  i = 0;
  while (cmd[i] && isspace(cmd[i]))
    i++;

  if (!isalpha(cmd[i]))	{	/* Malformed command */
    printf("blank line?\n");
    return 0;
  }

  /* copy the command word, translating to lower case */
  while (cmd[i] && isalpha(cmd[i])) {
    command[cmdLen] = isupper(cmd[i]) ? tolower(cmd[i]) : cmd[i];
    cmdLen++;
    if (cmdLen >= WORD_LEN) {	/* would overflow */
      printf("Command too long\n");
      return 0;
    }
    i++;
  }
  command[cmdLen] = 0;

  /* skip spaces */
  while (cmd[i] && isspace(cmd[i]))
    i++;

  if (!cmd[i]) {		/* Ooops no argument */
    printf("No argument in %s?\n", cmd);
    return 0;
  }

  while (cmd[i]) {
    if (isupper(cmd[i])) {
      argument[argLen] = tolower(cmd[i]);
    } else {
      argument[argLen] = cmd[i];
    }
    argLen++;
    i++;
    if (argLen >= WORD_LEN) {	/* would overflow */
      printf("Argument too long\n");
      return 0;
    }
  }

  argument[argLen] = 0;

  /* look for the command in the list. */
  for (i = 0, cnt = 0; Commands[i].c_name; i++) {
    if (strncmp(command,Commands[i].c_name,cmdLen) == 0) {
      cnt++;
      cmdNo = i;
    }
  }

  /* cnt = 0 -> no cmd; cnt > 1 -> ambiguous prefix */
  if (cnt == 0) {
    printf("No such command as %s\n", command);
    return 0;
  } else if (cnt > 1) {
    printf("%s is ambiguous\n", command);
    return 0;
  }

  switch (Commands[cmdNo].c_type) {
  case TYPE_ONOFF:
  case TYPE_LIST:
    argNo = findArg(argument, Commands[cmdNo].c_opt);
    if (argNo == -1) {
      printf("%s is an invalid argument for %s\n", argument, command);
      return 0;
    }
    sprintf(buffer,"%s%s", Commands[cmdNo].c_prefix,
	    Commands[cmdNo].c_opt[argNo].dest);
    break;
  case TYPE_RANGE:
    argNo = toInt(argument);
    if (argNo < 0) {
      printf("argument %s is not numeric.\n", argument);
      return 0;
    } else if (argNo > 255) {
      printf("argument %s is out of range (> 255).\n", argument);
      return 0;
    } else {
      sprintf(buffer,"%s%02x", Commands[cmdNo].c_prefix, argNo);
    }
    break;
  case TYPE_FREQ:
    freq = toDouble(argument);
    if (freq < 0) {
      printf("argument %s is not numeric.\n", argument);
      return 0;
    } else if (freq < 0.050 || freq > 1300.0) {
      printf("argument %s is out of range (0.050 - 1300.0).\n", argument);
      return 0;
    } else {
      sprintf(buffer,"%s%010.0f", Commands[cmdNo].c_prefix, freq*1000*1000);
    }
    break;
  case TYPE_BAND:
    return 0;
    break;
  default:
    return 0;			/* can't happen */
  }

  retval = malloc(strlen(buffer)+1);
  if (retval)
    strcpy(retval,buffer);
  return retval;

}
